package jwars;

import java.io.IOException;

import jwars.net.DefaultServerEventHandler;
//import jwars.net.JWarIOHandler;
import jwars.net.Server;
import jwars.control.networkevents.NetworkEvents;
import jwars.synch.Timer;
import jwars.synch.TimerListener;
import jwars.Util;
import jwars.test.JWProperties;
import jwars.model.WarModel;

/**
 * This class represents a complete JWars server. It will manage a Server
 * and the acceptance of clients joining. Finally it will instruct the Server
 * to send updates with regular intervals.
 */
public class ServerSession
    implements TimerListener{

  //Latency measured in frames
  public final int LATENCY;
  private int time;
  private Server warServer;

  /**
   * Create a ServerSession using the argument Timer to send updates
   * at regular intervals.
   *
   * @param timer Timer
   */
  public ServerSession(Timer timer){
    System.out.println("Loading Server Session");
    String latencyString = Util.SV_PROPS.getProperty(JWProperties.Server.latency.name());
    double latencySeconds = Double.parseDouble(latencyString);
    int latency = WarModel.WORLD.secondsToFrames(latencySeconds);
    if(latency < 2){
      latency = 2;
      System.out.println("Warning: latency very small, using latency 2 instead");
    }
    LATENCY = latency;
    timer.addTimerListener(this);
    try{
      this.warServer = new Server(NetworkEvents.getJWarsProtocol(), Util.TIMER,
          new DefaultServerEventHandler(), Util.getPort(), LATENCY);
    } catch(IOException ex){
      ex.printStackTrace();
      System.out.println("Unable to create server!");
    }
  }

  /**
   * If the Timer is not running, we still want to send updates at the same
   * interval
   */
  public void reportStasis(){
    poll();
  }

  /**
   * Send update if the set time since last update has been reached.
   * @param frameCount int
   * @param isLate boolean
   */
  public void update(int frameCount, boolean isLate){
    poll();
  }

  /**
   * Instructs the Server to cease listening for additional clients.
   */
  public void stopListening(){
    warServer.stopListening();
  }

  /**
   * Send a keep-alive instruction two times per interval of length LATENCY
   */
  private void poll(){
    if(time % (LATENCY / 2) == 0)
      warServer.sendKeepAliveInstruction();
    time++;
  }

}
