package jwars.control;

import java.util.ArrayList;
import java.util.List;

import jwars.model.Unit;
import jwars.model.WorldListener;

/**
 * Manages a single unit selection.
 */
public class UnitSelectionModel implements WorldListener {

  private List<UnitSelectionListener> listeners = new ArrayList<UnitSelectionListener>();

  private Unit markedUnit = null;

  /**
   * Notifies listeners that a new unit has been selected.
   * @param newUnit Unit
   * @param oldUnit Unit
   */
  private void unitSelected(Unit newUnit, Unit oldUnit){
    for(UnitSelectionListener listener : listeners){
      listener.unitSelected(newUnit, oldUnit);
    }
  }

  /**
   * Selects, or marks, the specified unit.
   * @param u the unit. This parameter may be null.
   */
  public void markUnit(Unit u){
    if(u != null && !u.isSelectable()){
      return;
    }
    Unit oldSelection = markedUnit;
    markedUnit = u; //Do this *before* calling unitSelected so markedUnit()
    //behaves consistently
    unitSelected(u, oldSelection);
  }

  /**
   * Adds a UnitSelectionListener such that it will be notified of any
   * changes in the unit selection.
   * @param listener UnitSelectionListener
   */
  public void addListener(UnitSelectionListener listener){
    listeners.add(listener);
  }

  /**
   * Get the currently selected unit.
   * @return the unit
   */
  public Unit markedUnit(){
    return markedUnit;
  }

  /**
   * Remove a UnitSelectionListener such that it will no longer receive
   * unit selection notifications.
   * @param listener UnitSelectionListener
   */
  public void removeListener(UnitSelectionListener listener){
    listeners.remove(listener);
  }

  /**
   * Invoked when a unit is now onboard. This method does nothing.
   * @param u Unit
   */
  public void unitNowOnBoard(Unit u){
  }

  /**
   * Invoked whena unit is destroyed. If that unit is the selection, the
   * new selection will be null.
   * @param u Unit
   */
  public void unitDestroyed(Unit u){
    if(u == markedUnit){
      markUnit(null);
    }
  }

}
