package jwars.control;

import java.util.ArrayList;

import jwars.model.map.CoordinateSystem;

/**
 * Class for managing the focus point on the map. A ViewPort is a rectangular
 * area on the map which is considered in focus. The location of a ViewPort
 * can be changed, and listeners for such changes can be registered and will
 * be notified of such changes.
 *
 * The ViewPort location refers to the coordinates of the central point of this
 * ViewPort (and not, for instance, the upper left corner).
 */
public class ViewPort{

  private int portWidth, portHeight;
  private int currentLocationX, currentLocationY;
  private final CoordinateSystem coords;

  private ArrayList<ViewPortListener> listeners =
      new ArrayList<ViewPortListener>();

  /**
   * Creates a ViewPort using the specified CoordinateSystem. All ViewPort
   * locations will pertain to the specified coordinate system.
   *
   * @param coords the native CoordinateSystem of this ViewPort
   * @param startX the start location x coordinate of this ViewPort
   * @param startY the start location y coordinate of this ViewPort
   * @param portWidth the ViewPort width
   * @param portHeight the ViewPort height
   */
  public ViewPort(CoordinateSystem coords, int startX, int startY,
                  int portWidth, int portHeight){
    this.coords = coords;
    this.portWidth = portWidth;
    this.portHeight = portHeight;
    setLocation(startX, startY);
  }

  /**
   * Returns the x coordinate of this ViewPorts location
   * @return int
   */
  public int getCurrentX(){
    return currentLocationX;
  }

  /**
   * Returns the y coordinate of this ViewPorts location
   * @return int
   */
  public int getCurrentY(){
    return currentLocationY;
  }

  /**
   * Sets the location of this ViewPort such that the upper left corner is at
   * the specified location
   * @param x int
   * @param y int
   */
  private void setLocationWithoutCompensation(int x, int y){
    x = (x < 0 ? 0 : x); //Restrict sign and magnitude of arguments
    y = (y < 0 ? 0 : y); //Nicer than 16 lines of if-else
    x = (x + portWidth >= coords.WIDTH ? coords.WIDTH - portWidth - 1 : x);
    y = (y + portHeight >= coords.HEIGHT ? coords.HEIGHT - portHeight - 1 : y);

    this.currentLocationX = x;
    this.currentLocationY = y;
    for(ViewPortListener v : listeners){
      v.reportLocationChange(x, y);
    }

  }

  /**
   * Sets the size of this ViewPort.
   * @param width int
   * @param height int
   */
  public void setSize(int width, int height){
    portWidth = width;
    portHeight = height;
  }

  /**
   * Sets the location of the center of this ViewPort.
   * @param x int
   * @param y int
   */
  public void setLocation(int x, int y){
    x -= portWidth / 2; //Compensation for origin in corner
    y -= portHeight / 2;
    setLocationWithoutCompensation(x, y);
  }

  /**
   * Nudges the location of this ViewPort by the specified amounts in x and y
   * directions. Values may be negative.
   * @param dx int
   * @param dy int
   */
  public void nudge(int dx, int dy){
    setLocationWithoutCompensation(this.currentLocationX + dx,
        this.currentLocationY + dy);
  }

  /**
   * Adds a ViewPortListener which will henceforth be notified of location
   * changes
   * @param listener ViewPortListener
   */
  public void addListener(ViewPortListener listener){
    listeners.add(listener);
  }

  /**
   * Listener interface for ViewPort location changes.
   */
  public static interface ViewPortListener{

    /**
     * Invoked when the ViewPort location is changed, if this ViewPortListener
     * is registered with that ViewPort.
     * @param x the new x coordinate of the center of the ViewPort
     * @param y the new y coordinate of the center of the ViewPort
     */
    public void reportLocationChange(int x, int y);

  }

}
