package jwars.gui;

import java.io.PrintWriter;
import java.io.Writer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;

/**
 * Console is a small combined chat and command console. Console consists
 * of a text field into which the user can enter input, and a text pane in which
 * output is displayed.
 *
 * Input will be relayed to a TextParser.
 */
public class Console
    extends JPanel{

  private JTextPane outputPane = new JTextPane();
  private JTextField inputField = new JTextField();
  private StyledDocument log = outputPane.getStyledDocument();
  private Component defaultFocusedComponent = null;
  private PrintWriter out = new PrintWriter(new ConsoleWriter(), true);
  private TextParser parser;

  private JScrollPane scrollPane =
      new JScrollPane(outputPane, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                      JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

  private Style defaultStyle = outputPane.addStyle("",null);
  private Style inputStyle = outputPane.addStyle("",null);

  /**
   * Creates a console which does not use a TextParser.
   */
  public Console(){
    this(new TextParser(){
      public String parse(String s){
        return s;
      }
    });
  }

  /**
   * Create a Console which returns focus to the argument Component
   * after having been used. Input Strings will be passed to the argument
   * parser.
   *
   * @param defaultFocusedComponent Component
   */
  public Console(TextParser parser){
    this.parser = parser;
    this.setLayout(new BorderLayout());

    inputField.setFont(new Font("Monospaced", Font.PLAIN, 12));
    outputPane.setFont(new Font("Monospaced", Font.PLAIN, 12));

    outputPane.setEditable(false);

    outputPane.setRequestFocusEnabled(false);
    inputField.setRequestFocusEnabled(false);

    ConsoleListener listener = new ConsoleListener();
    inputField.addActionListener(listener);
    inputField.addMouseListener(listener);

    this.add(scrollPane, BorderLayout.CENTER);
    this.add(inputField, BorderLayout.SOUTH);

    generateStyles();
  }

  /**
   * Puts this Console in focus such that its text field receives keyboard
   * input.
   */
  public void enableTyping(){
    inputField.setEnabled(true);
    inputField.requestFocusInWindow();
  }

  /**
   * Print a line to this Console's text pane.
   *
   * @param text the String
   * @param style the desired Style
   */
  public void println(String text, Style style){
    print(text+'\n', style);
  }

  /**
   * Print a String to this Console's text pane.
   * @param text the String
   * @param style the desired Style
   */
  public void print(String text, Style style){
    try{
      log.insertString(log.getLength(), text, style);
    } catch(BadLocationException ex){
      ex.printStackTrace();
    }
    outputPane.setCaretPosition(log.getLength());
  }

  /**
   * Creates and stores some styles internally
   */
  private void generateStyles(){
    StyleConstants.setForeground(defaultStyle, Color.black);
    StyleConstants.setForeground(inputStyle, Color.black);
  }

  /**
   * Clears this Console's text area.
   */
  public void clear(){
    outputPane.setText("");
  }

  /**
   * Get the
   * @return JTextPane
   */
  public JTextPane getOutputPane(){
    return outputPane;
  }

  public JTextField getInputField(){
    return inputField;
  }

  /**
   * Get a PrintWriter which writes to the output pane of this Console.
   * @return PrintWriter the PrintWriter
   */
  public PrintWriter getPrintWriter(){
    return out;
  }

  /**
   * After having received input, this Console will return focus to the default
   * focused component, provided that such an opponent has been set.
   *
   * @param comp Component
   */
  public void setDefaultFocusedComponent(Component comp){
    this.defaultFocusedComponent = comp;
  }

  /**
   * Get the default focused Component.
   * @return Component
   */
  public Component getDefaultFocusedComponent(){
    return this.defaultFocusedComponent;
  }

  /**
   * Set the color with which input text is echoed, provided that it is echoed
   * (as specified by the TextPbarser's return values) to the text pane.
   * @param color Color
   */
  public void setInputColor(Color color){
    System.out.println("input color set");
    StyleConstants.setForeground(inputStyle, color);
  }

  /**
   * Set the default color with which text will be displayed in the console.
   * Text entered through the PrintWriter returned by the getPrintWriter()
   * method will have this color.
   * @param color Color
   */
  public void setDefaultColor(Color color){
    System.out.println("def color set");
    StyleConstants.setForeground(defaultStyle, color);
  }

  /**
   * Sets the text background of this Console, i.e. the background of both
   * input field and text pane.
   * @param color Color
   */
  public void setTextBackgroundColor(Color color){
    inputField.setBackground(color);
    outputPane.setBackground(color);
  }

  /**
   * Utility class which allows interfacing with this Console's output pane
   * by means of Writers.
   */
  private class ConsoleWriter
      extends Writer{

    /**
     * Print the specified part of argument char[] to console
     * @param chars char[]
     * @param offset int
     * @param len int
     */
    public void write(char[] chars, int offset, int len){
      Console.this.print(new String(chars, offset, len), defaultStyle);
    }

    public void flush(){} //Flushes automatically

    public void close(){} //No one will discover that we didn't implement this

  }

  /**
   * Manages ActionEvents fired by this Console's input field.
   */
  private class ConsoleListener extends MouseAdapter
      implements ActionListener{

    public void actionPerformed(ActionEvent e){
      String s = inputField.getText();
      inputField.setText("");
      if(defaultFocusedComponent != null){
        inputField.setEnabled(false);
        defaultFocusedComponent.requestFocusInWindow();
      }
      if(!s.equals("")){
        //println(s, inputStyle);
        String response = parser.parse(s);
        if(response != null){
          println(response, inputStyle);
        }
      }
    }

    /**
     * Allow the user to start typing by clicking on the input field
     * @param e MouseEvent
     */
    public void mouseClicked(MouseEvent e){
      enableTyping();
    }

  }

}

