package jwars.gui;

import java.io.PrintWriter;
import java.util.List;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.MouseEvent;
import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.ListCellRenderer;

import jwars.control.UnitSelectionListener;
import jwars.data.ArmourTable;
import jwars.event.dynamic.DEvent;
import jwars.model.Formation;
import jwars.model.Moveable;
import jwars.model.Unit;
import jwars.model.WarModel;
import jwars.model.units.InfantrySquad;
import jwars.model.units.Vehicle;
import jwars.model.weapons.Weapon;

/**
 * This panel contains information about the selected unit.
 */
public class InfoPanel
    extends JPanel implements UnitSelectionListener{

  private JList list = new JList(new DefaultListModel());
  private ListCellRenderer renderer = new CellRenderer();

  private InfantryStatusPanel infantryPanel = new InfantryStatusPanel();
  private TankStatusPanel tankPanel = new TankStatusPanel();
  private JPanel left = new JPanel();
  private JPanel right = new JPanel();

  /**
   * Creates an Infopanel.
   */
  public InfoPanel(){
    super(new BorderLayout());
    list.setCellRenderer(renderer);
    list.setRequestFocusEnabled(false);
    list.setFocusable(false);
    list.setOpaque(false);
    list.setDragEnabled(false);
    list.setBorder(BorderFactory.createTitledBorder(""));
    left.setLayout(new FlowLayout(FlowLayout.LEFT));
    right.setLayout(new FlowLayout(FlowLayout.RIGHT,0,0));
    left.add(list);
    //left.setBorder(BorderFactory.createTitledBorder(""));
    add(left, BorderLayout.WEST);
    add(right, BorderLayout.EAST);
    jwars.control.Control.UNIT_SELECTION.addListener(this);
    list.addMouseListener(new java.awt.event.MouseAdapter(){
      public void mousePressed(MouseEvent mouseEvent){
        int i = list.locationToIndex(mouseEvent.getPoint());
        list.setSelectedIndex(i);
        Object o = list.getSelectedValue();
        if(o instanceof Unit && ((Unit)o).isAlive()){
          jwars.control.Control.UNIT_SELECTION.markUnit((Unit)o);
        } else if(o instanceof Weapon){
          writeWeaponInfo((Weapon)o);
        }
      }
    });
    unitSelected(null, null);
  }

  /**
   * Updates the displayed data according to new unit selection.
   *
   * @param newSelection Unit
   * @param lastSelection Unit
   */
  public void unitSelected(Unit  newSelection, Unit lastSelection){
    writeWeaponInfo(null);
    DefaultListModel model = (DefaultListModel)list.getModel();
    model.clear();
    if(newSelection == null){
      list.setBorder(BorderFactory.createTitledBorder(""));
      right.removeAll();
      return;
    }
    String name = newSelection.toString();
    list.setBorder(BorderFactory.createTitledBorder(name));
    List objects;
    if(newSelection instanceof Moveable){
      Moveable m = (Moveable)newSelection;
      objects = m.getWeaponModel().getWeaponList();
      updateInformation(m);
    } else if(newSelection instanceof Formation){
      Formation f = (Formation)newSelection;
      objects = f.getSubUnitList();
      right.removeAll();
    } else{
      objects = new java.util.ArrayList<String>(); //Not reached
    }
    for(Object o : objects){
      model.addElement(o);
    }
    revalidate();
    //revalidate();
    //repaint();
  }

  /**
   * Writes weapon data to console
   * @param w Weapon
   */
  private void writeWeaponInfo(Weapon w){
    if(w == null){
      return;
    }
    PrintWriter out = jwars.Util.out;
    out.println("Weapon\t" + w);
    out.println("Ammo\t" + w.CAT.type.name());
    out.print("Range\t");
    out.println(Double.toString(WarModel.WORLD.toMetres((int)w.CAT.range)) +
                "m");
    out.print("Eff.range\t");
    out.println(Double.toString(WarModel.WORLD.toMetres((int)w.CAT.
        effectiveRange)) + "m");
    out.print("Steel pen.\t");
    out.println(Integer.toString((int)w.CAT.baseSteelPenetration) + "mm");
    out.print("Firepower\t");
    out.println(Integer.toString((int)w.CAT.killValue));
    out.print("Reload\t");
    out.println(Double.toString(WarModel.WORLD.framesToSeconds(w.CAT.reloadTime)) +
                "s");
  }

  /**
   * Updates display according to the specified Moveable.
   * @param m Moveable
   */
  private void updateInformation(Moveable m){
    right.removeAll();
    if(m instanceof jwars.model.units.Vehicle){
      right.add(tankPanel);
      tankPanel.updateInfo((Vehicle)m);
    } else if(m instanceof jwars.model.units.InfantrySquad){
      infantryPanel.updateInfo((InfantrySquad)m);
      right.add(infantryPanel);
    }
  }

  /**
   * Panel designed to display information about infantry squads.
   */
  private class InfantryStatusPanel
      extends JPanel{

    private JLabel label = new JLabel("Men: "), value = new JLabel();
    private InfantrySquad selection;
    private DEvent update = new DEvent(){
      public void execute(){
        if(InfantryStatusPanel.this.isVisible() && selection != null){
          value.setText(Integer.toString(selection.getStrength()));
        }
        WarModel.WORLD.waitFor(this, WarModel.WORLD.secondsToFrames(2.0));
      }
    };


    public InfantryStatusPanel(){
      setLayout(new GridLayout(1, 2));
      update.execute();
      add(label);
      add(value);
    }

    /**
     * Update contents according to the argument InfantrySquad.
     * @param is InfantrySquad
     */
    public void updateInfo(InfantrySquad is){
      this.selection = is;
      value.setText(Integer.toString(is.getStrength()));
    }

  }

  /**
   * Panel designed to display information about vehicles.
   */
  private class TankStatusPanel
      extends JPanel{

    private ArmourDisplayPanel
        hull = new ArmourDisplayPanel("Hull armour"),
        turret = new ArmourDisplayPanel("Turret armour");

    /**
     * Creates a TankStatusPanel
     */
    public TankStatusPanel(){
      setLayout(new GridLayout(2, 1, 0, 0));
      add(hull);
      add(turret);
    }

    /**
     * Updates the contents according to the specified vehicle.
     * @param vehicle Vehicle
     */
    public void updateInfo(Vehicle vehicle){
      hull.update(vehicle.getArmourModel().hull);
      turret.update(vehicle.getArmourModel().turret);
    }

  }

  /**
   * Displays information about a particular ArmourTable.
   */
  private class ArmourDisplayPanel
      extends JPanel{

    private JLabel
        frontLabel = new JLabel("Front"),
        sideLabel = new JLabel("Side"),
        rearLabel = new JLabel("Rear"),
        front = new JLabel(),
        side = new JLabel(),
        rear = new JLabel();

    public ArmourDisplayPanel(String title){
      setLayout(new GridLayout(3,2,0,0));
      setBorder(BorderFactory.createTitledBorder(title));
      add(frontLabel);
      add(front);
      add(sideLabel);
      add(side);
      add(rearLabel);
      add(rear);
    }

    /**
     * Update contents according to the specified ArmourTable
     * @param armour ArmourTable
     */
    public void update(ArmourTable armour){
      front.setText(armourToString(armour.FRONT, armour.ANGLE_FRONT));
      side.setText(armourToString(armour.SIDE, armour.ANGLE_SIDE));
      rear.setText(armourToString(armour.REAR, armour.ANGLE_REAR));
    }

    /**
     * Converts thickness and angles in as floats to a human readable String.
     *
     * @param thickness armour thickness in millimetres
     * @param angle armour slope in millimetres
     * @return the String representation
     */
    public String armourToString(float thickness, float angle){
      int t = (int)thickness;
      int a = (int)Math.toDegrees(angle);
      return Integer.toString(t) + " mm @ " + a + "\u00B0";
    }

  }

  /**
   * Cellrenderer for the list of sub-units or weapons.
   */
  private class CellRenderer
      extends JLabel implements ListCellRenderer{
    public CellRenderer(){
      setOpaque(true);
    }

    public Component getListCellRendererComponent(JList list, Object value,
        int index, boolean isSelected, boolean cellHasFocus){
      setText(value.toString());
      Color color = Color.black;
      if(value instanceof Unit){
        if(!((Unit)value).isAlive()){
          color = color.red;
        }
      }
      setForeground(color);
      return this;
    }
  }

}
