package jwars.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import jwars.Util;
import jwars.control.Control;
import jwars.control.InputManager;
import jwars.model.Unit;
import jwars.model.WarModel;

/**
 * This is the frame which JWars runs in.
 */
public class WarFrame
    extends JFrame{

  private WarPanel warPanel;
  private MapPanel mapPanel;
  private Console console;
  private InputManager wControl;
  //private Community community;

  private int scrollSpeed = 3 * WarModel.TIME_PER_FRAME/2;

  /**
   * Creates a WarFrame using the specified InputManager. This will register
   * a number of commands with that InputManager.
   *
   * @param wControl InputManager
   */
  public WarFrame(InputManager wControl){
    super(Util.NAME_AND_VERSION);
    this.wControl = wControl;
    this.console = new Console(wControl);
    console.setDefaultFocusedComponent(this);
    this.addKeyListener(wControl);
    this.addFocusListener(wControl);
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //this.addMouseMotionListener(new Listener());
    addPanels();
    setUndecorated(Util.isFullScreen() || Util.useSpecialATIHack);
    if(Util.isFullScreen()){
      setSize(Util.DISPLAY_SIZE_X(), Util.DISPLAY_SIZE_Y());
    } else {
      pack();
    }
    setResizable(false);
    addKeyBindings();
  }

  /**
   * Add different panels to the contentPane and set their sizes depending
   * on the screen size defined in Util
   */
  private void addPanels(){
    Container contentPane = this.getContentPane();
    contentPane.setLayout(new BorderLayout());

    int chunkX = Util.DISPLAY_SIZE_X() / 4;
    int chunkY = Util.DISPLAY_SIZE_Y() / 4;
    //Note: may produce trouble if not divisible by 4

    this.warPanel = new WarPanel(3 * chunkX, 3 * chunkY);
    //Util.setOut(console.getPrintWriter());
    //wControl.setOut(console.getPrintWriter());
    console.setBorder(BorderFactory.createEtchedBorder());
    JComponent inputField = console.getInputField();
    JComponent outputPane = console.getOutputPane();
    //Prevent console from stealing focus
    inputField.setEnabled(false);
    outputPane.setFocusable(false);

    //Initialize panels
    JComponent
        westPanel = new JPanel(/*new FlowLayout(FlowLayout.LEFT, 0, 0)*/),
        centerPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0)),
        infoPanel = new InfoPanel(),
        overViewPanel = new OverViewPanel();
        //mapPanelContainer = new JPanel(new FlowLayout(FlowLayout.CENTER)),

    westPanel.setLayout(new BoxLayout(westPanel, BoxLayout.Y_AXIS));

    this.mapPanel = new MapPanel(warPanel,chunkX);

    //Set preferred sizes in terms of window size in pixels

    warPanel.setPreferredSize(new Dimension(3 * chunkX, 3 * chunkY));
    //Presently done in WarPanel constructor, too

    infoPanel.setPreferredSize(new Dimension(3 * chunkX - 5 * chunkX / 4,
                                             chunkY));
    console.setPreferredSize(new Dimension(5 * chunkX / 4, chunkY));
    overViewPanel.setPreferredSize(new Dimension(chunkX, 4 * chunkY - chunkX));
    //mapPanel.setPreferredSize(new Dimension(chunkX, chunkX));
    //square; done in constructor of mapPanel, too;

    centerPanel.setPreferredSize(new Dimension(3 * chunkX, 4 * chunkY));
    westPanel.setPreferredSize(new Dimension(chunkX, 4 * chunkY));

    //Add panels to each other
    centerPanel.add(warPanel);
    centerPanel.add(infoPanel);
    centerPanel.add(console);
    westPanel.add(overViewPanel);

    JPanel mapPanelContainer = new JPanel();
    mapPanelContainer.setLayout(new BoxLayout(mapPanelContainer,
        BoxLayout.X_AXIS));

    mapPanelContainer.add(Box.createHorizontalGlue());
    mapPanelContainer.add(mapPanel);

    westPanel.add(Box.createVerticalGlue());
    westPanel.add(mapPanelContainer);
    contentPane.add(centerPanel, BorderLayout.CENTER); //Main view, center
    contentPane.add(westPanel, BorderLayout.WEST);
  }

  /**
   * This method will be deprecated later. Only the map panel needs manual
   * rendering.
   */
  public void drawPanels(){
    mapPanel.render();
  }

  /**
   * Returns the Console displayed in this WarFrame.
   * @return Console
   */
  public Console getConsole(){
    return console;
  }

  /**
   * Returns the WarPanel displayed in this WarFrame.
   * @return WarPanel
   */
  public WarPanel getWarPanel(){
    return warPanel;
  }

  /**
   * Registers a number of key bindings with the InputManager of this WarFrame.
   */
  private void addKeyBindings(){
    InputManager.KeyTrackEvent e = new InputManager.KeyTrackEvent(wControl.defaultQueue){
      public void execute(){
        console.enableTyping();
      }
    };
    wControl.put("typecommand"/*unusable obviously*/, KeyEvent.VK_ENTER, e);
    e = new InputManager.KeyTrackEvent(wControl.defaultQueue){
      public void execute(){
        System.exit(0);
      }
    };
    wControl.put("clear", new InputManager.KeyTrackEvent(wControl.defaultQueue){
      public void execute(){
        console.clear();
      }
    });
    wControl.put("exit", KeyEvent.VK_ESCAPE, e);
    wControl.put("quit", e);
    wControl.put("super", KeyEvent.VK_BACK_SPACE, new SelectSuperUnitEvent());
    e = new InputManager.KeyTrackEvent(wControl.defaultQueue){

      public void execute(){
        Util.out.println(""+java.util.Calendar.getInstance().getTime());
      }

    };
    wControl.put("time", e);

    scrollUp = new ScrollEvent(0,1);
    scrollDown = new ScrollEvent(0,-1);
    scrollLeft = new ScrollEvent(-1,0);
    scrollRight = new ScrollEvent(1,0);

    wControl.put("nudgeDown", KeyEvent.VK_DOWN, scrollUp);
    wControl.put("nudgeUp", KeyEvent.VK_UP, scrollDown);
    wControl.put("nudgeLeft", KeyEvent.VK_LEFT, scrollLeft);
    wControl.put("nudgeRight", KeyEvent.VK_RIGHT, scrollRight);
  }

  private ScrollEvent
      scrollUp,// = new ScrollEvent(0,1),
      scrollDown,// = new ScrollEvent(0,-1),
      scrollLeft,// = new ScrollEvent(-1,0),
      scrollRight;// = new ScrollEvent(1,0);

  /**
   * A command which selects the super unit of the currently selected unit.
   * This might be refactored elsewhere.
   */
  private class SelectSuperUnitEvent extends InputManager.KeyTrackEvent{

    public SelectSuperUnitEvent(){
      super(wControl.defaultQueue);
    }

    public void execute(){
      Unit markedUnit = Control.UNIT_SELECTION.markedUnit();
      if(markedUnit != null)
        warPanel.markUnit(markedUnit.getSuperUnit());
    }

  }

  /**
   * Event which is used for zooming around the map (in one step).
   */
  private class ZoomEvent extends InputManager.KeyTrackEvent{

    public int x, y;

    public ZoomEvent(){
      super(wControl.defaultQueue);
    }

    protected void execute(){
      warPanel.VIEWPORT.setLocation(x, y);
    }

  }

  /**
   * Event which is used for scrolling around the map (continuously).
   */
  private class ScrollEvent extends InputManager.KeyTrackEvent{

    private int dx, dy;

    public ScrollEvent(int dx, int dy){
      super(wControl.defaultQueue);
      this.dx = dx;
      this.dy = dy;
    }

    protected void execute(){
      super.execute();
      warPanel.VIEWPORT.nudge(scrollSpeed*dx, scrollSpeed*dy);
    }

  }

  /**
   * Listener which will be used to allow mouse scrolling of the battlefield.
   * This is disabled as it doesn't work correctly, for example there can be
   * trouble with the mouse exiting the frame (implement fix for this someday).
   */
  private class Listener implements MouseMotionListener{

    public void mouseDragged(MouseEvent e){}
    public void mouseMoved(MouseEvent e){
      int x = e.getX();
      int y = e.getY();

      System.out.println("x "+x);
      System.out.println("y "+y);

      if(x < 10){
        scrollLeft.enqueue();
      } else if(x > WarFrame.this.getContentPane().getWidth()-10){
        scrollRight.enqueue();
      }
      if(y < 10){
        scrollUp.enqueue();
      } else if(y > WarFrame.this.getContentPane().getHeight()-10){
        scrollDown.enqueue();
      }
    }

  }

}
