package jwars.model;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jwars.event.EventQueue;
import jwars.event.TimedEvent;
import jwars.model.ai.Attackable;
import jwars.model.units.Team;
import jwars.model.ai.UnitAI;
import jwars.net.IOHandler;

/**
 * Unit is the base class of all entities in JWars. Units are organized into
 * a tree structure, in which nodes represent military formations such as
 * companies or platoons, and leaves represent single bottom-level units
 * such as individual vehicles.
 *
 * This class provides an abstraction of these two concepts with the
 * objective of allowing the user to control forces without worrying if
 * those forces consist of single entities or complete armies at
 * a time.
 */
public class Unit implements Attackable {

  final static int bitsPerNestingLevel = 5;
  final static int maxUnits = (1<<bitsPerNestingLevel) - 1;
  final static int treeDepth = 6;
  //The above variables are used to index units in the tree.

  //Not final, because if a Formation gets a new leader we want location
  //to point towards that Unit's location. However this might be a problem
  //with AI objects which store the original reference...
  //AIs generally only store the location of Moveables, not Formations,
  //so this is not a problem presently
  protected Coord location = new Coord();
  protected final Coord destination = new Coord();
  protected double theta;
  protected final int radius;

  boolean onBoard = false;
  boolean alive = true;

  int id; //Uniquely identifies this Unit in the unit tree

  Unit superUnit;
  final List<Unit> subUnits = new ArrayList<Unit>(0);

  protected UnitAI controller = UnitAI.PASSIVE_AI;

  private String name;

  protected World world;

  public Unit(String name, int radius){
    this.name = name;
    this.radius = radius;
  }

  public void update(){
    Unit u;
    for(int i = subUnits.size()-1; i >= 0; i--){
      u = subUnits.get(i);
      if(u.isAlive()){
        u.update();
      }
    }
  }

  public void addUnit(Unit unit){
    if(unit.onBoard)
      throw new RuntimeException("Unit already on board!");
    if(subUnits.size() >= maxUnits){
      throw new RuntimeException("Too many sub units! Maximum is " + maxUnits);
    }
    subUnits.add(unit);
    unit.superUnit = this;
    unit.id = (this.id << bitsPerNestingLevel) + subUnits.size();
    unit.revalidateIds();
    //unit.newSuperior(this);
  }

  public void addUnits(List<? extends Unit> units){
    for(Unit u : units){
      addUnit(u);
    }
  }

  public void dirtify(){
    if(!isAlive()){
      return;
    }
    for(Unit u : subUnits){
      u.dirtify();
    }
  }

  private void revalidateIds(){
    for(int i = 0; i < subUnits.size(); i++){
      Unit u = subUnits.get(i);
      u.id = (this.id << bitsPerNestingLevel) + i+1;
      u.revalidateIds();
    }
  }

  void checkAllIds(World world){
    if(world.getUnit(this.id) != this){
      String errorMsg =
          "Error in Unit hierarchy!\n"+
          "ID of "+this+ " corresponds to "+world.getUnit(this.id);
      throw new RuntimeException(errorMsg);
    }
    for(Unit u : subUnits){
      u.checkAllIds(world);
    }
  }

  /**
   * Add this Unit to the specified World at the specified coordinates.
   * @param world World
   * @param x int
   * @param y int
   */
  public void addToWorld(int x, int y, boolean addChildren){
    if(this.superUnit == null)
      throw new RuntimeException("Unit ("+this+") not in hierarchy!");
    if(this.onBoard){
      throw new RuntimeException("Unit already on board!");
    }
    if(!this.superUnit.onBoard){
      throw new RuntimeException("Cannot add "+this+
                                 " to world: super Unit not on board!");
    }
    if(!alive){
      throw new RuntimeException(this+" is dead, must be offboard");
    }
    this.location.x = x;
    this.location.y = y;
    this.world = superUnit.world;
    this.onBoard = true;
    //System.out.println("<Unit> Added to world: "+this+" x="+x+" y="+y);
    world.unitRegistered(this);
    if(addChildren){
      addChildrenToWorld();
    }
  }

  protected void addChildrenToWorld(){
    for(Unit u : subUnits){
      if( u.isAlive() )
        u.addToWorld(u.getX(), u.getY(), true);
    }
  }

  private boolean hasLiveSubUnit(){
    for(int i = subUnits.size()-1; i>=0; i--){
      if(subUnits.get(i).isAlive()){
        return true;
      }
    }
    return false;
  }

  protected void removeFromWorld(){
    //should be made public so removal and destruction are not equivalent
    if(hasLiveSubUnit()){
      throw new RuntimeException("Cannot remove unit: sub units still onboard");
    }
    world.unitRemoved(this);
    this.onBoard = false;
    world = null;
  }

  public void destroy(){
    this.alive = false;
    setAI(UnitAI.PASSIVE_AI);
    superUnit.reportSubUnitDestroyed(this);
    removeFromWorld();
  }

  public boolean isAlive(){
    return alive;
  }

  public boolean isOnBoard(){
    return onBoard;
  }

  /**
   * This method is invoked when a sub unit of this Unit is removed. The
   * default implementation is empty.
   * @param unit the Unit which has been removed
   */
  protected void reportSubUnitDestroyed(Unit unit){
    //System.out.println("sub unit reported removed!!");
    if(!hasLiveSubUnit()){
      destroy();
    }
  }

  /**
   * Retrieve a subUnit of this Unit, or this Unit.
   * @param index index of sub Unit.
   * @return Unit sub Unit number index, or this if index is -1.
   */
  public Unit getUnit(int index){
    return index == -1 ? this : subUnits.get(index);
  }

  public List<Unit> getSubUnitList(){
    return subUnits;
  }

  /*protected void newSuperior(Unit superior){

  }*/

  /**
   * Get the number of sub-units of this Unit.
   * @return the sub-unit count
   */
  public int getSubUnitCount(){
    return subUnits.size();
  }

  /**
   * Get the super Unit of this Unit.
   * @return Unit the super Unit.
   */
  public Unit getSuperUnit(){
    return superUnit;
  }

  public Team getTeam(){
    return superUnit == null ? null : superUnit.getTeam();
  }

  /**
   * Returns whether argument unit is an ancestor of this Unit in the Unit tree.
   * If this Unit is equal to other, this method will return true.
   * @param other Unit
   * @return boolean
   */
  public boolean subordinateOf(Unit other){
    if(other == null)
      return false;
    boolean result = false;
    Unit superior = this;
    while(superior != null && !result){
      result = (superior == other);
      superior = superior.superUnit;
    }
    return result;
  }

  /**
   * Returns whether this Unit is an ancestor of the argument in the Unit tree.
   * @param other Object
   * @return boolean
   */
  public boolean superiorOf(Object other){
    return other instanceof Unit && ((Unit)other).subordinateOf(this);
  }

  public UnitAI getAI(){
    return controller;
  }

  /**
   * Set the AI model of this Unit.
   * @param controller the new AI
   */
  public void setAI(UnitAI controller){
    this.controller = controller;
  }

  /**
   * Get the ID of this Unit
   * @return the ID
   */
  public int getId(){
    return id;
  }

  /**
   * Get the x-coordinate of this Unit
   * @return the x-coordinate
   */
  public int getX(){
    return location.x;
  }

  /**
   * Get the y-coordinate of this Unit
   * @return the y-coordinate
   */
  public int getY(){
    return location.y;
  }

  public double getAngle(){
    return theta;
  }

  protected void setAngle(double theta){
    this.theta = theta;
  }

  /**
   * Get the radius of this Unit
   * @return the radius
   */
  public int getRadius(){
    return radius;
  }

  public void setDestination(int x, int y){
    this.destination.set(x,y);
  }

  public String toString(){
    return name();
    //return super.toString() + " ID = " + id;
  }

  public boolean isSelectable(){
    return isAlive();
  }

  /**
   * This method should be called when the user issues an order directly to this
   * Unit. The order will be sent to other computers on the network before it
   * is carried out. This will set the destination of this Unit to the specified
   * location.
   * @param x x-coordinate of destination
   * @param y y-coordinate of destination
   */
  public void applyMove(int x, int y){
    moveHandler.orderMove(this, x, y, UnitMoveEvent.SIMPLE_MOVE);
  }

  public void appendMove(int x, int y){
    moveHandler.orderMove(this, x, y, UnitMoveEvent.APPEND_MOVE);
  }

  public void prependMove(int x, int y){
    moveHandler.orderMove(this, x, y, UnitMoveEvent.PREPEND_MOVE);
  }

  public void attack(int x, int y, int radius){
    attackHandler.orderAttack(this, x, y, radius);
  }

  public void sendStopCommand(){
      stopHandler.orderStop(this);
  }

  /*public void attack(Unit target){

  }*/

  public int getDestinationX(){
    return destination.x;
  }

  public int getDestinationY(){
    return destination.y;
  }

  public String name(){
    return name;
  }

  /**
   * Constructs a Unit which is suitable for use as a root of a unit tree.
   * @return the Unit
   */
  public static Unit createContainerUnit(World world){
    Unit u = new Unit("Top-level Container Unit", 0){

      public void addToWorld(int x, int y, boolean addChildren){
        this.onBoard = true;
      }

      public boolean isSelectable(){
        return false;
      }

    };
    u.world = world;
    return u;
  }

  /**
   * This Serializer is responsible for sending movement instructions across
   * network.
   */
  public final static MoveOrderHandler moveHandler = new MoveOrderHandler();
  public final static AttackOrderHandler attackHandler = new AttackOrderHandler();
  public final static StopOrderSerializer stopHandler = new StopOrderSerializer();

  private static class StopOrderSerializer extends IOHandler<Object> {

    int id;

    public void orderStop(Unit u){
      this.id = u.id;
      try{
        send(null);
      } catch(Exception ex){
        ex.printStackTrace();
      }
    }

    protected void read(DataInput input, int timeStamp) throws IOException{
      int id = input.readInt();
      new UnitStopEvent(WarModel.WORLD.getUnit(id), timeStamp);
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      output.writeInt(input.readInt());
    }

    protected void write(Object t, DataOutput output) throws IOException{
      output.writeInt(id);
    }

  }

  private static class AttackOrderHandler extends IOHandler<int[]>{

    private int[] attackOrderData = new int[4];

    public void orderAttack(Unit u, int x, int y, int r){
      attackOrderData[0] = u.id;
      attackOrderData[1] = x;
      attackOrderData[2] = y;
      attackOrderData[3] = r;
      try{
        send(attackOrderData);
      } catch(IOException ex){
        ex.printStackTrace();
      }
    }

    protected void read(DataInput input, int timeStamp) throws IOException{
      Unit u = WarModel.WORLD.getUnit(input.readInt());
      int x = input.readInt();
      int y = input.readInt();
      int type = input.readInt();
      new UnitAttackEvent(u, x, y, type, timeStamp);
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      for(int i = 0; i < attackOrderData.length; i++){
        output.writeInt(input.readInt());
      }
    }

    protected void write(int[] t, DataOutput output) throws IOException{
      for(int i = 0; i < attackOrderData.length; i++){
        output.writeInt(attackOrderData[i]);
      }
    }

  }

  /**
   * Class for sending move orders across a network
   */
  private static class MoveOrderHandler extends IOHandler<int[]>{

    //Stores the ID and destination temporarily
    private int[] moveOrderData = new int[4];

    /**
     * Send a move order for the specified Unit to the specified location
     * @param u Unit to be moved
     * @param x x-coordinate of target location
     * @param y y-coordinate of target location
     */
    public void orderMove(Unit u, int x, int y, int type){
      moveOrderData[0] = u.id;
      moveOrderData[1] = x;
      moveOrderData[2] = y;
      moveOrderData[3] = type;
      try{
        send(moveOrderData);
      } catch(IOException ex){
        ex.printStackTrace();
      }
    }

    protected void write(int[] data, DataOutput output) throws IOException{
      for(int i = 0; i < data.length; i++){
        output.writeInt(data[i]);
      }
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      output.writeInt(input.readInt());
      output.writeInt(input.readInt());
      output.writeInt(input.readInt());
      output.writeInt(input.readInt());
    }

    protected void read(DataInput input, int timeStamp) throws IOException{
      Unit u = WarModel.WORLD.getUnit(input.readInt());
      int x = input.readInt();
      int y = input.readInt();
      int type = input.readInt();
      new UnitMoveEvent(u, x, y, type, timeStamp);
    }

  }

  /**
   * Get the EventQueue used to manage moving units.
   * @return the EventQueue
   */
  public static EventQueue getUnitMoveQueue(){
    return networkQueue;
  }

  private final static EventQueue networkQueue = new EventQueue(0);

  private static class UnitAttackEvent extends TimedEvent {

    //uses same queue. Actually this queue should be defined elsewhere, named NetworkQueue.
    //public final static EventQueue attackQueue = networkQueue;

    private final Unit u;
    private final int x, y, radius;

    public UnitAttackEvent(Unit u, int x, int y, int radius, int time){
      super(networkQueue);
      this.u = u;
      this.x = x;
      this.y = y;
      this.radius = radius;
      enqueue(time);
    }

    public void execute(){
      u.getAI().orderAttack(x, y, radius);
    }

  }

  private static class UnitStopEvent extends TimedEvent {

    private Unit u;

    public UnitStopEvent(Unit u, int time){
      super(networkQueue);
      this.u = u;
      enqueue(time);
    }

    protected void execute(){
      u.controller.orderHalt(null);
      u.controller.orderHoldFire();
    }

  }

  /**
   * This class represents a move order waiting to be executed.
   */
  private static class UnitMoveEvent extends TimedEvent{

    private final Unit u;
    private final int x, y;
    private final int type;

    private final static int SIMPLE_MOVE = 1;
    private final static int PREPEND_MOVE = 2;
    private final static int APPEND_MOVE = 3;

    public UnitMoveEvent(Unit u, int x, int y, int type, int time){
      super(networkQueue);
      //System.out.println("<Unit> Move event created at time "+jwars.Util.TIMER.getTime());
      this.u = u;
      this.x = x;
      this.y = y;
      this.type = type;
      if(type != SIMPLE_MOVE && type != PREPEND_MOVE && type != APPEND_MOVE){
        throw new IllegalArgumentException(
            "Unrecognized movement event modifier: "+type);
      }
      enqueue(time);
      //System.out.println("<Unit> Move order enqueued at time "+jwars.Util.TIMER.getTime());
    }

    public void execute(){
      switch(type){
        case SIMPLE_MOVE:
          u.controller.orderMove(null, x,y);
          break;
        case PREPEND_MOVE:
          u.controller.prependMove(null, x,y);
          break;
        case APPEND_MOVE:
          u.controller.appendMove(null, x,y);
          break;
      }
      //System.out.println("<Unit> Executing move order " +this+" at time "+jwars.Util.TIMER.getTime());

    }

  }

}
