package jwars.model.collision;


import jwars.model.*;

/**
 * In order to register an entity in the CollisionDetector, that entity must
 * possess a CollisionProperties set.
 *
 * CollisionProperties specify whether Colliders can move, whether they are
 * massive (two Colliders can move through each other if neither is massive),
 *
 * CollisionProperties can be subclassed in order to specify custom shapes
 * of the Collider.
 */
public class CollisionProperties{

  CollisionTile generalLocation;
  int height = 0;
  boolean mobile;
  boolean massive;
  boolean ghost = false;
  final Collider collider;
  //Two units that are both
  //Ghosts do not collide with anything

  boolean marked = false;
  //The collision detector uses this variable for internal 'voodoo'.
  //Specifically for avoiding counting elements twice if they appear in
  //multiple collision tiles.

  /**
   * Creates a CollisionProperties belonging to the specified Collider which is
   * massive or mobile as specified.
   *
   * @param collider Collider
   * @param massive boolean
   * @param mobile boolean
   */
  public CollisionProperties(Collider collider, boolean massive, boolean mobile){
    this.massive = massive;
    this.mobile = mobile;
    this.collider = collider;
  }

  /**
   * Returns whether this Collider can collide with the argument Collider.
   * @param other the other Collider
   * @return whether this Collider can collide with the other.
   */
  public boolean mayCollide(CollisionProperties other){
    //No collision if: at least one Collider is a ghost
    //---or---
    //neither Collider is massive
    return !(ghost || other.ghost || !(massive || other.massive));
  }

  public void setMassive(boolean massive){
    this.massive = massive;
  }

  /**
   * Override this to implement custom shapes. This method is invoked when
   * a Collider's bounding circle overlaps another Collider's bounding circle
   * and returns whether a collision has happened.
   *
   * The default implementation returns true, meaning that colliders are
   * circular per default.
   *
   * @param other the Collider whose bounding circle overlaps that of this
   * Collider
   * @return whether a collision has occured
   */
  public boolean checkCollide(Collider other, int x, int y){
    return true;
  }

  /*public void setMobile(boolean mobile){
    this.mobile = mobile;
  }*/

}
