package jwars.model.map;

import java.util.Random;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;

import jwars.Util;
import jwars.model.map.generator.FloatBuffer;
import jwars.model.map.generator.TerrainGenerator;

import jwars.model.collision.*;

/**
 * This class contains static methods for generating maps.
 */
public class MapFactory{

  public final static int MAX_FORESTATION = 6;

  public static TileMap<CollisionTile> createCollisionMap(CoordinateSystem mother, int scale){
    int width = mother.WIDTH/scale;
    int height = mother.HEIGHT/scale;
    CollisionTile[][] tiles = new CollisionTile[width][height];
    for(int i = 0; i < tiles.length; i++){
      for(int j = 0; j < tiles[0].length; j++){
        tiles[i][j] = new CollisionTile();
      }
    }
    return new TileMap<CollisionTile>(mother, tiles);
  }

  /**
   * Creates a TileMap of TerrainTiles with the correct number of tiles
   * according to the specified scale and supercoordinate system.
   * @param mother CoordinateSystem
   * @param scale int
   * @return TileMap
   */
  public static TileMap<TerrainTile> createTerrainMap(CoordinateSystem mother,
      int scale){
    int width = mother.WIDTH / scale;
    int height = mother.HEIGHT / scale;

    TerrainTile[][] tiles =
        new TerrainTile[width][height];
    BufferedImage[] grassTypes = new BufferedImage[16];

    for(int i=0; i < grassTypes.length; i++){
      grassTypes[i] = Util.createImage(scale,scale);
      createGrass(grassTypes[i],i);
      grassTypes[i] = Util.recreate(grassTypes[i]);
    }

    TerrainGenerator generator = new TerrainGenerator(width, height, 5L);
    FloatBuffer forestationMap = generator.generateTrees(0.46f, MAX_FORESTATION, 2);
    int treeVariations = 6;
    BufferedImage[][] forestImages = generateForestTileImages(scale, treeVariations);
    Random rand = new Random(0);
    int tileCount = 0;
    for(int i = 0; i < tiles.length; i++){
      for(int j = 0; j < tiles.length; j++){
        int forestTypeIndex = (int)forestationMap.get(tileCount++);
        if(forestTypeIndex > MAX_FORESTATION){
          System.out.println("max forestation exceeded");
          forestTypeIndex = MAX_FORESTATION - 1;
        }
        tiles[i][j] = new TerrainTile(
            grassTypes[rand.nextInt(grassTypes.length)],
            forestImages[forestTypeIndex][rand.nextInt(treeVariations)],
            i * scale, j * scale,
            forestTypeIndex);
      }

    }
    return new TileMap<TerrainTile>(mother, tiles);
  }

  private static BufferedImage[][] generateForestTileImages(int scale, int variations){
    BufferedImage[][] forestTypes = new BufferedImage[MAX_FORESTATION][variations];
    for(int i = 0; i < forestTypes.length; i++){
      for(int j = 0; j < variations; j++){
        forestTypes[i][j] = Util.createImage(scale, scale,
            java.awt.Transparency.BITMASK);
        createTrees(forestTypes[i][j], j, scale/7, 2*scale/5, i * 3-1);
        forestTypes[i][j] = Util.recreate(forestTypes[i][j]);
      }
    }
    return forestTypes;
  }

  private static void createGrass(/*Buffered*/Image img, long seed){
    //Graphics2D g2d = img.createGraphics();
    Graphics g = img.getGraphics();
    int width = img.getWidth(null);
    int height = img.getHeight(null);
    Random rand = new Random(seed);
    for(int i=0; i < width; i++){
      for(int j=0; j < height; j++){
        g.setColor(new Color(
                     18+rand.nextInt(128),
                     96+rand.nextInt(72),
                     6 +rand.nextInt(72)
                     ));
        drawPoint(i, j, g);
      }
    }
    g.dispose();
  }

  private static void drawPoint(int x, int y, Graphics g){
    g.drawLine(x, y, x, y);
  }

  /**
   * Vegetation-relevant stuff
   */

  private final static Color treeColor = Color.green.darker().darker().darker();
  private final static Color transparent = new Color(0,0,0,0);

  private static void createTrees(BufferedImage im, long seed, int minDiameter, int maxDiameter, int count){
    Random rand = new Random(seed);
    int w = im.getWidth(null);
    int h = im.getHeight(null);
    Graphics g = im.getGraphics();
    for(int i = 0; i < count; i++){
      int diam = rand.nextInt(maxDiameter-minDiameter) + minDiameter;
      int x0 = rand.nextInt(w-diam)+diam/2;
      int y0 = rand.nextInt(h-diam)+diam/2;
      float baseGreenColor = 0.18f + rand.nextFloat()*0.15f;
      for(int j = 0; j < diam*diam*2; j++){
        Color c = new Color(0, baseGreenColor+rand.nextFloat()/3.f, 0);
        g.setColor(c);
        float r = diam * rand.nextFloat() / 2 + 2;
        float theta = (float)(2*Math.PI * rand.nextFloat());
        int x = x0 + (int)(r*Math.cos(theta));
        int y = y0 + (int)(r*Math.sin(theta));
        g.drawLine(x,y,x,y);
      }
    }
    g.dispose();
  }

}
