package jwars.model.map;

import java.util.Iterator;

public class TileMap<T> extends CoordinateSystem{

  final T[][] tiles;

  /**
   * Constructs a TileMap inside the specified CoordinateSystem using the
   * specified array as tiles. If the dimensions of the array are incompatible
   * with the super coordinate system, an exception is thrown.
   * @param superSystem supersystem of this TileMap
   * @param tiles T[][] elements of this TileMap
   */
  public TileMap(CoordinateSystem superSystem, T[][] tiles){
    super(superSystem, checkArraySize(superSystem, tiles));
    this.tiles = tiles;
    assert(super.HEIGHT == tiles.length);
    assert(super.WIDTH == tiles[0].length);
  }

  /**
   * Used for constructing subclasses
   * @param map TileMap
   */
  public TileMap(TileMap<T> map){
    super(map.superSystem, map.scale);
    this.tiles = map.tiles;
  }

  /**
   * Get a specific tile
   * @param i horizontal position of tile
   * @param j vertical position of tile
   * @return T tile located at argument coordinates
   */
  public T getTile(int i, int j){
    return tiles[i][j];
  }

  /**
   * Get the tile containing the point (x,y) in the main coordinate system.
   * If this TileMap is the main coordinate system, this is equivalent to
   * getTile(x,y).
   * @param x 1st coordinate in main coordinate system
   * @param y 2nd coordinate in main coordinate system
   * @return the tile at (x,y)
   */
  public T getTileAt(int x, int y){
    return tiles[x/scale][y/scale];
  }

  /**
   * Checks that the argument array is compatible with the mother
   * CoordinateSystem and returns the scaling factor of this TileMap
   * @param tiles Object[][]
   * @return int
   */
  private static int checkArraySize(CoordinateSystem superSystem, Object[][] tiles){
    int width = tiles.length;
    int height = tiles[0].length;
    int xScale = superSystem.WIDTH/width;
    int yScale = superSystem.HEIGHT/height;
    if(xScale != yScale){
    //if(superSystem.WIDTH%width != 0 || superSystem.HEIGHT%height != 0 ||
    //   xScale != yScale){
      try{
        throw new Exception("Illegal tile array dimensions");
      }
      catch(Exception ex){
        ex.printStackTrace();
      }
    }
    assert(xScale == yScale);
    return xScale;
  }

  /**
   * Calls the operate method of the specified TileOperation on all
   * tiles contained within the bounds given.
   * @param tileOp TileOperation
   * @param startX int
   * @param startY int
   * @param endX int
   * @param endY int
   */
  public void performOp(TileOperation<T> tileOp, int startX, int startY, int endX, int endY){
    for(int i = startX; i <= endX; i++){
      for(int j = startY; j <= endY; j++){
        tileOp.operate(tiles[i][j]);
      }
    }
  }

  /**
   * Calls the operate method of the specified TileOperation on all
   * tiles contained within the bounds given by the argument int[],
   * using the first two values as upper left corner and the two latter as
   * lower right corner. Further elements will be ignored. If the array has
   * length less than 4, this method will throw an
   * ArrayIndexOutOfBoundsException.
   *
   * @param tileOp TileOperation
   * @param bounds int[]
   */
  public void performOp(TileOperation<T> tileOp, int[] bounds){
    performOp(tileOp, bounds[0], bounds[1], bounds[2], bounds[3]);
  }

  public static interface TileOperation<T>{

    public void operate(T tile);

  }

}
