package jwars.model.vision;

import jwars.event.Event;
import jwars.test.Link;

public class ObserverModel
    extends Link<ObserverModel>{

  final ObservationEnvironment env;
  final Event event;
  final Observer subject;
  ObserverTile generalLocation;
  private boolean isAlive = true;

  private boolean hasBeenSpotted = false;
  private int timeOfLastSpotting;

  public ObserverModel(ObservationEnvironment env, Observer subject){
    super.setValue(this);
    this.env = env;
    this.event = new ObservationEvent();
    this.subject = subject;
  }

  protected void performCompleteCheck(int dt){
    updateLocation();
    //System.out.println("completecheck");
    hasBeenSpotted = isStillVisible();
    //now invisible unless it is getting discovered again
    //The environment knows how the tile layout looks and will kindly
    //invoke the check methods of this object appropriately
    env.traverseSurroundingTiles(this, dt);
  }

  protected void performTileCheck(ObserverTile tile, int dt){
    //Traverse tile, perform check on each element
    for(Link<ObserverModel> link : tile){
      ObserverModel obs = link.value();
      if(!subject.sharesVision(obs.subject)){
        boolean spotted = performCheck(obs, dt);
        if(spotted){
          obs.hasBeenSpotted = true;
          obs.timeOfLastSpotting = env.timer.getTime();
        }
      }
    }
  }

  private boolean isStillVisible(){
    return hasBeenSpotted && (env.timer.getTime() - timeOfLastSpotting <= env.relaxationTime);
  }

  private boolean isOrWillSoonBeHidden(){
    return env.timer.getTime() - timeOfLastSpotting > env.relaxationTime*0.7f;
  }

  protected void updateLocation(){
    ObserverTile tile = env.obsMap.getTileAt(subject.getX(), subject.getY());
    if(tile != generalLocation){
      remove();
      generalLocation = tile;
      tile.addLast(this);
    }
  }

  public boolean hasBeenSpotted(){
    return this.hasBeenSpotted;
  }

  void deactivate(){
    remove();
    generalLocation = null;
    isAlive = false;
    hasBeenSpotted = false;
    timeOfLastSpotting = 0;
  }

  /*public boolean sharesVision(ObserverModel other){
    if(other == this){
      return true;
    }
    return true; //implement later
  }*/

  protected boolean performCheck(ObserverModel other, int dt){
    float dx = (float)other.subject.getX()-this.subject.getX();
    float dy = (float)other.subject.getY()-this.subject.getY();
    float distSqr = dx*dx + dy*dy;
    float visibilitySqr = (float)env.maxVisibility * env.maxVisibility;
    return (distSqr < visibilitySqr) && subject.visionCheck(other.subject, (int)Math.sqrt(distSqr), dt);
  }

  private class ObservationEvent
      extends Event{

    private int lastCheckTime = 0;

    public ObservationEvent(){
      super(env.getObservationQueue());
    }

    protected void execute(){
      if(!isAlive){
        return;
      }
      int currentTime = env.timer.getTime();
      int timeChange = currentTime - lastCheckTime;
      lastCheckTime = env.timer.getTime();
      performCompleteCheck(timeChange);
      enqueue();
    }

  }

}
