package jwars.model.weapons;

import java.util.List;
import java.util.ArrayList;
import jwars.data.WeaponCategory;
import jwars.event.Event;
import jwars.event.EventQueue;
import jwars.model.WarModel;
import jwars.model.Coord;
import jwars.test.explosions.Explosion;
import jwars.model.collision.Collider;
import jwars.model.Moveable;
import jwars.event.dynamic.DEvent;

public class Weapon{

  //private ReloadEvent reloadEvent;
  private boolean isLoaded = true;
  public final WeaponCategory CAT;
  private ReloadListener listener;
  private List<Collider> hitList = new ArrayList<Collider>();
  private int salvoSize = 1;
  private ImpactDistCalculator distCalculator =
      new DefaultImpactDistCalculator(this);

  private final DEvent reloadEvent = new DEvent(){
    public void execute(){
      isLoaded = true;
      listener.reportReloaded(Weapon.this);
    }
  };

  public Weapon(WeaponCategory category){
    this.CAT = category;
    //reloadEvent = new ReloadEvent(queue);
  }

  public boolean fire(int srcX, int srcY, int dstX, int dstY){
    if(!isLoaded){
      throw new RuntimeException("Cannot fire weapon which is not loaded!");
    }
    float dx = dstX - srcX;
    float dy = dstY - srcY;
    float distSqr = dx*dx + dy*dy;
    if(distSqr > CAT.range * CAT.range){
      return false; //out of range
    }
    isLoaded = false;
    int reloadTime = CAT.reloadTime*(20+WarModel.WORLD.rand.nextInt(8))/24;
    WarModel.WORLD.waitFor(reloadEvent, reloadTime);
    //reloadEvent.enqueue();
    calculateImpactLocation(srcX, srcY, dstX, dstY, (int)Math.sqrt(distSqr));
    return true;
  }

  protected void calculateImpactLocation(int srcX, int srcY, int dstX, int dstY, int distance){
    int hitX = dstX;
    int hitY = dstY;

    if(CAT.type != ShellType.bullet){
      float deviationAngle, deviationRadius;
      deviationAngle = (float)(WarModel.WORLD.rand.nextFloat() * 2 * Math.PI);
      deviationRadius = (float)distCalculator.impactDistFromTarget(distance);

      int deviationX = (int)(deviationRadius * Math.cos(deviationAngle));
      int deviationY = (int)(deviationRadius * Math.sin(deviationAngle));

      hitX += deviationX;
      hitY += deviationY;
    }
    calculateDamage(srcX, srcY, hitX, hitY);
  }

  protected void calculateDamage(int srcX, int srcY, int hitX, int hitY){
    hitList.clear();
    WarModel.WORLD.detector.getOverlapping(hitList, hitX, hitY, (int)CAT.splashRange);

    float travellingRange = (int)Math.hypot(hitX-srcX, hitY-srcY);
    float travellingAngle = (float)Math.atan2(hitY-srcY, hitX-srcX);

    weaponFired(srcX, srcY, travellingAngle);

    Moveable m;
    float impactDistance;
    for(Collider c : hitList){
      if(c instanceof Moveable){
        m = (Moveable)c;
        impactDistance = (float)Math.hypot(hitX-m.getX(), hitY-m.getY());
        //We have to add/subtract pi to the impact angle because an impact of 0
        //is defined to be on the front of the entity hit
        m.getArmourModel().reportImpact(this, hitX, hitY, impactDistance,
            travellingRange, (float)(travellingAngle-m.getAngle()-Math.PI));
      }
    }
    weaponHit(srcX, srcY, hitX, hitY);
  }

  protected void weaponFired(int srcX, int srcY, float travellingAngle){
    //draw muzzle flash, play sound, for example
    //actually it's a bit difficult because srcX and srcY do not point to the
    //muzzle.
  }

  protected void weaponHit(int srcX, int srcY, int hitX, int hitY){
    WarModel.WORLD.addDynamicSprite(new Explosion(CAT.explosionType, hitX, hitY));
  }

  public boolean isLoaded(){
    return isLoaded;
  }

  void setReloadListener(ReloadListener listener){
    this.listener = listener;
  }

  private class ReloadEvent extends Event {

    public ReloadEvent(EventQueue queue){
      super(queue);
    }

    protected void execute(){
      isLoaded = true;
      listener.reportReloaded(Weapon.this);
    }

  }

  public void setSalvoSize(int size){
    this.salvoSize = size;
  }

  public int getSalvoSize(){
    return salvoSize;
  }

  public String toString(){
    return CAT.name;
  }

  public void setImpactDistCalculator(ImpactDistCalculator calc){
    this.distCalculator = calc;
  }

  public static interface ImpactDistCalculator {
    public float impactDistFromTarget(int range);
  }

  public static class DefaultImpactDistCalculator implements ImpactDistCalculator {

    protected Weapon weapon;

    public DefaultImpactDistCalculator(Weapon weapon){
      this.weapon = weapon;
    }

    public float impactDistFromTarget(int range){
      double badness = range / weapon.CAT.effectiveRange;
      float unluckiness = WarModel.WORLD.rand.nextFloat();
      float deviation = (float)(0.2f * range * badness * unluckiness);
      return Math.abs(deviation);
    }

  }

}
