package jwars.synch;

import java.lang.reflect.InvocationTargetException;
import java.util.Vector;

import java.awt.EventQueue;

/**
 * The Timer class is used to deliver updates at regular intervals.
 * Individual updates are approximately equidistant unless the updates cannot
 * be executed in the alotted time, in which case the Timer will attempt
 * to execute them as fast as possible at the risk of starvation of other
 * threads. The framerate
 * is self-adjusting such that it will never diverge from some internal timer
 * such as System.currentTimeMillis().
 *<p></p>
 * Care should be taken that sufficient time is alotted for updates to be
 * executed. During the update operation it is possible to check whether the
 * Timer is late, and thus reschedule nonessential updates.
 * <p></p>
 * The methods of this class are unsynchronized. This means that this Timer
 * should be manipulated only through the update method of the TimerListener
 * it updates.
 */
public class Timer{

  private final int DELAY; //delay between updates in milliseconds
  private final TimerThread thread = new TimerThread();
  private final Watch watch; //The clock used for synchronizing
  private final Vector<TimerListener> listeners =
      new Vector<TimerListener>(); //Listener for updates
  private boolean isRunning = false; //False when paused or suspended
  private boolean isPaused = false; /* Used internally for telling the
              difference between pause and suspend states */

 private int frameCount = 0; //Amount of updates performed
  //private boolean isRunning = false;

  //Synchronization variables
  private long startTime;
  private long pausedTime = 0;
  private int laxness = 4;

  /**
   * Constructs a Timer object which will invoke the
   * update methods of every registered TimerListener every delayMillis
   * milliseconds as
   * specified by the default Watch, which uses System.nanoTime().
   *
   * @param delayMillis delay in milliseconds between ticks
   * @param listener listener the target of this Timer
   */
  public Timer(int delayMillis){
    this(new NanoWatch(), delayMillis);
  }

  /**
   * Constructs a Timer object using argument Watch, which will invoke the
   * update method of argument TimerListener every delayMillis milliseconds as
   * specified by the Watch.
   *
   * @param watch the Watch of this Timer
   * @param delayMillis delay in milliseconds between ticks
   * @param listener the target of this Timer
   */
  public Timer(Watch watch, int delayMillis){
    this.watch = watch;
    this.DELAY = delayMillis;
  }

  /**
   * Starts this Timer. Calling this method twice on the same Timer object
   * will result in an IllegalThreadStateException being thrown.
   */
  public void start(){
    isRunning = true;
    startTime = watch.currentTimeMillis();
    thread.start();
  }

  /**
   * Add a TimerListener to this Timer.
   * @param listener TimerListener to be added
   */
  public void addTimerListener(TimerListener listener){
    listeners.add(listener);
  }

  /**
   * Pauses this Timer. No more updates will be performed until unpause() is
   * invoked. After being unpaused, the Timer will update at the normal rate,
   * i.e. will not try to catch up with the Watch. Subsequent calls without
   * unpause() will do nothing. While paused, the Timer will call the
   * reportStasis() method of the target TimerListener with the same frequency
   * with which it would ordinarily call update().
   */
  public void pause(){
    isRunning = false;
    isPaused = true;
  }

  /**
   * Suspends this Timer. While the Timer is suspended, no more updates will
   * be performed until unpause() is invoked. After being unpaused, the Timer
   * will catch up with the reading of the Watch as quickly as possible, i.e.
   * consider itself "late" if it has been suspended for long enough.
   *
   * This method can be used if the application must still be running at full
   * speed, but cannot presently continue due to lacking input.
   *
   * While paused, the Timer will call the
   * reportStasis() method of the target TimerListener with the same frequency
   * with which it would ordinarily call update().
   */
  public void suspend(){
    isRunning = false;
  }

  /**
   * Unpauses this Timer if it has been suspended or paused. If both paused and
   * suspended, the Timer will catch up with the current reading of the Watch,
   * minus the time after the pause() method was invoked.
   */
  public void unpause(){
    isRunning = true;
    isPaused = false;
  }

  /**
   * Returns the number of frames currently executed.
   * @return current frame count
   */
  public int getTime(){
    return frameCount;
  }

  /**
   * Returns true if and only if the current framecount is below what which
   * should have been executed as determined by the Watch.
   * @return whether the Timer is late
   */
  public boolean isLate(){
    return timeLeftMillis() < 0;
  }

  /**
   * Returns true if this Timer is currently running, false otherwise. Otherwise
   * means either that the Timer has not yet been started, is paused or
   * suspended.
   * @return whether the Timer is running
   */
  public boolean isRunning(){
    return isRunning;
  }

  /**
   * Returns true if and only if this Timer is currently paused.
   * @return whether the Timer is paused
   */
  public boolean isPaused(){
    return isPaused;
  }

  /**
   * Returns true if and only if this Timer is currently suspended.
   * @return whether the Timer is suspended.
   */
  public boolean isSuspended(){
    return!(isRunning || isPaused);
  }

  /**
   * Delay this Timer by specified number of milliseconds. This number may
   * be negative, forcing the Timer to keep up.
   * @param milliSeconds number of milliseconds to delay
   */
  public void delay(int milliSeconds){
    this.pausedTime += milliSeconds;
  }

  /**
   * Returns the number of milliseconds left before the Timer will be late
   * @return number of milliseconds left within current frame
   */
  public long timeLeftMillis(){
    long actualElapsedTime = watch.currentTimeMillis() - startTime - pausedTime;
    long targetElapsedTime = (frameCount+1) * DELAY;
    return targetElapsedTime - actualElapsedTime;
  }

  private void notifyListeners(long timeLeftMillis){
    for(TimerListener listener : listeners){
      listener.update(frameCount, timeLeftMillis < -laxness * DELAY);
    }
  }

  /**
   * The Thread used by the Timer class for sleeping.
   */
  private class TimerThread
      extends Thread{

    public TimerThread(){
      super("Synchronizer Thread");
    }

    public void run(){
      long timeLeftMillis = timeLeftMillis();
      while(true){
        if(isRunning){ //"Default" update
          frameCount++;
          try{
            EventQueue.invokeAndWait(invoker);
          } catch(InterruptedException ex){
            ex.printStackTrace();
          } catch(InvocationTargetException ex){
            ex.printStackTrace();
          }
          //notifyListeners(timeLeftMillis);
          //for(TimerListener listener : listeners){
          //  listener.update(frameCount, timeLeftMillis < 0);
          //}
        }
        else{ //Paused or suspended
          try{
            EventQueue.invokeAndWait(stasisInvoker);
          } catch(InterruptedException ex){
            ex.printStackTrace();
          } catch(InvocationTargetException ex){
            ex.printStackTrace();
          }
          //for(TimerListener listener : listeners){
          //  listener.reportStasis(); //Report not running
          //}
          if(isPaused)
            pausedTime += DELAY; //Compensation for lost time
        }
        timeLeftMillis = timeLeftMillis();
        try{
          if(timeLeftMillis > 0){
            sleep(timeLeftMillis);
          }
          else if(!isRunning && !isPaused){ //suspended, sleep no matter what
            sleep(DELAY);
          }
        }
        catch(InterruptedException ex){
          ex.printStackTrace();
        }
      }
    } //Method run

  } //Class TimerThread

  /**
   * Watch implementation which delegates to
   * System.currentTimeMillis()
   */
  private static class MilliWatch
      implements Watch{

    public long currentTimeMillis(){
      return System.currentTimeMillis();
    }

  }

  /**
   * Watch implementation which delegates to
   * System.nanoTime()
   */
  private static class NanoWatch
      implements Watch{

    public long currentTimeMillis(){
      return System.nanoTime() / 1000000;
    }

  }

  private Runnable invoker = new Runnable(){
    public void run(){
      notifyListeners(timeLeftMillis());
    }

  };

  private Runnable stasisInvoker = new Runnable(){
    public void run(){
      for(TimerListener listener : listeners){
        listener.reportStasis(); //Report not running
      }
    }
  };

  //private SwingInvoker invoker = new SwingInvoker();

  /*private class SwingInvoker implements Runnable {

    public void run(){
      notifyListeners(timeLeftMillis());
    }

  }*/

}
